(define (memoize size proc)
  "\
Return a version of PROC that memoizes its results, using a hash table of
size SIZE."
  (let ((table (make-hash-table size))
        (unique (list #f)))
    (define (memoized . args)
      (let* ((entry (hash-create-handle! table args unique)))
        (if (eq? (cdr entry) unique)
          (set-cdr! entry (apply proc args)))
        (cdr entry)))
    (set-procedure-property! memoized 'name (procedure-name proc))
    memoized))

(define (uniquifier size)
  "\
Return a procedure that returns a unique instance of each value passed to it,
using a hash table of size SIZE."
  (let ((table (make-hash-table size)))
    (define (uniquify obj)
      (car (hash-create-handle! table obj #f)))
    uniquify))
